#include "ESPAsyncWebServer.h"
#include <WiFi.h>
#include <Wire.h>
#include "HDC1080Sensor.h"
#include "CCS811Sensor.h"
#include <time.h>
//#include "src/ESPinfluxdb.h" // https://github.com/hwwong/ESP_influxdb // 14.04.2019
#include "http_static.h" // HTTP pages and JSON request templates
// ********************** Config **********************

// DeepSleep time – send data every 60 seconds
const int sleepTimeS = 60;

// WiFi Config
#define WiFi_SSID "Ischtar"
#define WiFi_Password "highfive"

// NTP conf
const char* ntpServer = "pool.ntp.org";
const long  gmtOffset_sec = 0;
const int   daylightOffset_sec = 3600;
struct tm timeinfo;

// Create AsyncWebServer object on port 80
AsyncWebServer server(80);

// Globals for HDC1080
HDC1080Sensor HDC1080_sensors;
double hdc1080_temp, hdc1080_humidity;
bool hdc1080_err;

// Globals for CCS811
CCS811Sensor CCS811_sensors;
uint16_t eco2, etvoc, errstat, raw;

// loop cycle
#define workCycle 60 //seconds
// ******************** Config End ********************

String readHDC1080Temperature() {
  // Sensor readings may also be up to 2 seconds 'old' (its a very slow sensor)
  if (isnan(hdc1080_temp)) {
    Serial.println("Failed to read from HDC1080 sensor!");
    return "--";
  }
  else {
    return String(hdc1080_temp);
  }
}

String readHDC1080Humidity() {
  // Sensor readings may also be up to 2 seconds 'old' (its a very slow sensor)
  if (isnan(hdc1080_humidity)) {
    Serial.println("Failed to read from HDC1080 sensor!");
    return "--";
  }
  else {
    return String(hdc1080_humidity);
  }
}

String processCCS811Error(err_t errstat) {
  if ( errstat == CCS811_ERRSTAT_OK_NODATA ) {
    Serial.println("CCS811: waiting for (new) data");
    return "loading";
  } else if ( errstat & CCS811_ERRSTAT_I2CFAIL ) {
    Serial.println("CCS811: I2C error");
    return "i2c error";
  } else {
    // Serial.print("CCS811: errstat="); Serial.print(errstat, HEX);
    // Serial.print("="); Serial.println( ccs811.errstat_str(errstat) );
    return "error";
  }
}

String readCCS811TVOC() {
  if( errstat==CCS811_ERRSTAT_OK ){
    return String(etvoc);
  } else {
    return processCCS811Error(errstat);
  }
}
String readCCS811ECO2() {
  if( errstat==CCS811_ERRSTAT_OK ){
    return String(eco2);
  } else {
    return processCCS811Error(errstat);
  }
}

String formatISO8601() {
    char timestamp[20]; // Buffer for timestamp

    // Format the time into ISO 8601 format
    strftime(timestamp, sizeof(timestamp), "%Y-%m-%dT%H:%M:%SZ", &timeinfo);

    // Convert the formatted timestamp to a String object
    return String(timestamp);
}


// Replaces placeholder in HTML template with real values
// SSR if you will
String processor(const String& var){
  if(var == "TEMPERATURE"){
    return readHDC1080Temperature();
  }
  else if(var == "HUMIDITY"){
    return readHDC1080Humidity();
  }
  else if(var == "TVOC"){
    return readCCS811TVOC();
  }
  else if(var == "ECO2"){
    return readCCS811ECO2();
  } else if(var == "TIMESTAMP"){
    return formatISO8601();
  }
  return String();
}

void connectToWiFi() {
  WiFi.mode(WIFI_STA);
  WiFi.begin(WiFi_SSID, WiFi_Password);

  Serial.println();
  Serial.print("Connecting to WiFi: ");
  Serial.print(WiFi_SSID);

  int attempts = 0;
  while (WiFi.status() != WL_CONNECTED && attempts < 10) {
    delay(500);
    Serial.print(".");
    attempts++;
  }

  if (WiFi.status() == WL_CONNECTED) {
    Serial.println("\nWiFi connected");
    Serial.print("IP address: http://");
    Serial.println(WiFi.localIP());
  } else {
    Serial.println("\nFailed to connect to WiFi");
    // Handle connection failure, e.g., retry or reset the ESP32
  }
}

void setup()
{
  Serial.begin(115200);
  delay(10);
  Serial.println("");

  connectToWiFi();
  delay(10);
  Serial.println("");
  // Config NTP
  configTime(gmtOffset_sec, daylightOffset_sec, ntpServer);

  // i2c
  Wire.begin();

  // humidity and temperature
  HDC1080_sensors.init();
  // eCO2 and eTVOC, temp and humidity needed to adjust values
  CCS811_sensors.init(&hdc1080_temp, &hdc1080_humidity);


  // Pages and JSONs
  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send_P(200, "text/html", http_static::index_html, processor);
  });
  server.on("/api/sensors.json", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send_P(200, "text/html", http_static::sensor_things_resp, processor);
  });

  // lightweight named endpoints
  server.on("/temperature", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send_P(200, "text/plain", readHDC1080Temperature().c_str());
  });
  server.on("/humidity", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send_P(200, "text/plain", readHDC1080Humidity().c_str());
  });
  server.on("/tvoc", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send_P(200, "text/plain", readCCS811TVOC().c_str());
  });
  server.on("/eco2", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send_P(200, "text/plain", readCCS811ECO2().c_str());
  });

  // Start server
  server.begin();
}

void loop()
{
  if (WiFi.status() != WL_CONNECTED) {
    Serial.println("WiFi connection lost. Reconnecting...");
    connectToWiFi();
  }

  if(!getLocalTime(&timeinfo)){
    Serial.println("Failed to obtain time");
    return;
  }

  Serial.print(&timeinfo, "[%M-%d-%Y--%H:%M:%S]: ");

  HDC1080_sensors.read_values(&hdc1080_temp, &hdc1080_humidity, &hdc1080_err);
  CCS811_sensors.read_values(&eco2, &etvoc, &errstat, &raw);

  Serial.print("H=");     Serial.print(readHDC1080Temperature());  Serial.print(" °C  ");
  Serial.print("T=");     Serial.print(readHDC1080Humidity());     Serial.print(" %  ");
  Serial.print("eco2=");  Serial.print(readCCS811ECO2());          Serial.print(" ppm  ");
  Serial.print("etvoc="); Serial.print(readCCS811TVOC());          Serial.print(" ppb  ");
  Serial.println("");
  // Wait
  delay(workCycle*1000);
}
